import unittest
import metaknowledge

class TestCitation(unittest.TestCase):
    def setUp(self):
        self.Cite = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1, DOI 0.1063/1.1695064")

    def test_citation_author(self):
        self.assertEqual(self.Cite.author, "JOHN D")

    def test_citation_year(self):
        self.assertEqual(self.Cite.year, "2015")

    def test_citation_journal(self):
        self.assertEqual(self.Cite.journal, "TOPICS IN COGNITIVE SCIENCE")

    def test_citation_v(self):
        self.assertEqual(self.Cite.V, "V1")

    def test_citation_p(self):
        self.assertEqual(self.Cite.P, "P1")

    def test_citation_DOI(self):
        self.assertEqual(self.Cite.DOI, "0.1063/1.1695064")

    def test_citation_id(self):
        self.assertEqual(self.Cite.getID(), "JOHN D, 2015, TOPICS IN COGNITIVE SCIENCE")

    def test_citation_str(self):
        self.assertEqual(str(self.Cite), "John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1, DOI 0.1063/1.1695064")

    def test_citation_extra(self):
        self.assertEqual(self.Cite.getExtra(), "TOPICS IN COGNITIVE SCIENCE, V1, P1")

    def test_citation_badDetection(self):
        self.assertTrue(metaknowledge.Citation("").bad)

    def test_citation_equality(self):
        c1 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, P1, DOI 0.1063/1.1695064")
        c2 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1")
        c3 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P2")
        self.assertTrue(c1 == self.Cite)
        self.assertTrue(c2 == self.Cite)
        self.assertFalse(c1 != c2)
        self.assertFalse(c3 == c1)

    def test_citation_hash(self):
        self.assertTrue(bool(hash(self.Cite)))
        self.assertTrue(bool(hash(metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1"))))
        self.assertTrue(bool(hash(metaknowledge.Citation("John D., 2015"))))

    def test_citation_badLength(self):
        c = metaknowledge.Citation("a, b")
        self.assertTrue(c.bad)
        self.assertEqual(str(c.error), "Too few elements")
        self.assertEqual(c.getExtra(),'B')
        self.assertEqual(c.author,'A')
        self.assertEqual(c.getID(),'A')

    def test_citation_badNumbers(self):
        c = metaknowledge.Citation("1 2, 2, 3, 4")
        self.assertTrue(c.bad)
        self.assertEqual(c.getID(), '1 2, 2')
        self.assertEqual(str(c.error), "Too many numbers")



"""Doc String for metaknowledge main
"""
from .record import Record, recordParser, BadISIRecord
from .citation import Citation, BadCitation, filterNonJournals
from .recordCollection import RecordCollection, isiParser
from .basicTest import btest
from .graphHelpers import write_edgeList, write_nodeAttributeFile, write_graph, read_graph, _ProgressBar, drop_edges, drop_nodesByDegree, drop_nodesByCount, graphStats
from .constants import VERBOSE_MODE, tagToFull, fullToTag, tagNameConverter, tagsAndNames, knownTagsList
from .blondel import blondel, modularity



import unittest
import metaknowledge
import os
import filecmp
import copy
import networkx as nx

class TestRecordCollection(unittest.TestCase):

    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")
        self.RCbad = metaknowledge.RecordCollection("metaknowledge/tests/badFile.isi")

    def test_iscollection(self):
        self.assertIsInstance(self.RC, metaknowledge.RecordCollection)
        self.assertEqual(repr(metaknowledge.RecordCollection()), "empty")
        self.assertTrue(self.RC == self.RC)

    def test_bad(self):
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/badFile.isi').bad)
        with self.assertRaises(TypeError):
            metaknowledge.RecordCollection('metaknowledge/tests/testFile.isi', extension= '.txt')
        self.assertTrue(self.RCbad + self.RC <= self.RC + self.RCbad)
        self.assertTrue(len(self.RCbad + self.RCbad) == 0)
        self.assertFalse(self.RCbad == self.RC)

    def test_badRecords(self):
        badRecs = self.RC.getBadRecords()
        self.assertTrue(badRecs <= self.RC)
        self.assertTrue(badRecs.pop().bad)
        self.RC.dropBadRecords()

    def test_dropJourn(self):
        RCcopy = copy.copy(self.RC)
        self.RC.dropNonJournals()
        self.assertEqual(len(self.RC), len(RCcopy) - 2)
        self.RC.dropNonJournals(invert = True)
        self.assertEqual(len(self.RC), 0)
        RCcopy.dropNonJournals(ptVal = 'B')
        self.assertEqual(len(RCcopy), 1)


    def test_getWOS(self):
        self.RC.dropBadRecords()
        R = self.RC.peak()
        l = len(self.RC)
        self.assertTrue(R, self.RC.getWOS(R.UT))
        self.assertEqual(len(self.RC), l)
        self.RC.dropWOS(R.UT)
        self.assertEqual(len(self.RC), l - 1)
        self.RC.getWOS(self.RC.peak().UT, drop = True)
        self.assertEqual(len(self.RC), l - 2)
        self.assertFalse(self.RC.getWOS(self.RC.pop().UT))
        with self.assertRaises(ValueError):
            self.RC.getWOS("asdfghjkjhgfdsdfghj")
            self.RC.dropWOS("asdfghjkjhgfdsdfghj")


    def test_directoryRead(self):
        self.assertEqual(len(metaknowledge.RecordCollection('.')), 0)
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/') >= self.RC)
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/', extension= '.txt') <= self.RC)

    def test_write(self):
        fileName = 'OnePaper2.isi'
        RC = metaknowledge.RecordCollection('metaknowledge/tests/' + fileName)
        RC.writeFile(fileName + '.tmp')
        RC.writeFile()
        self.assertTrue(filecmp.cmp('metaknowledge/tests/' + fileName, fileName + '.tmp'))
        self.assertTrue(filecmp.cmp('metaknowledge/tests/' + fileName, repr(RC)[:200] + '.isi'))
        os.remove(fileName + '.tmp')
        os.remove(repr(RC)[:200] + '.isi')

    def test_writeCSV(self):
        filename = "test_writeCSV_temporaryFile.csv"
        if os.path.isfile(filename):
            os.remove(filename)
        self.RC.writeCSV(filename, onlyTheseTags=['UT', 'PT', 'TI', 'AF','J9' ,'CR', 'pubMedID'], firstTags = ['CR', 'UT', 'J9', 'citations'], csvDelimiter = '∂', csvQuote='≠', listDelimiter= '«', longNames=True)
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(os.path.getsize(filename), 106373)
        os.remove(filename)
        self.RC.writeCSV(filename)
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(os.path.getsize(filename), 88201)
        os.remove(filename)

    def test_makeDict(self):
        d = self.RC.makeDict(onlyTheseTags = list(metaknowledge.tagsAndNames), longNames = True)
        self.assertEqual(len(d), 61)
        self.assertEqual(len(d['wosString']), len(self.RC))
        self.assertEqual(d['eISSN'][0], None)
        self.assertIsInstance(d['citations'], list)
        d = self.RC.makeDict(longNames = False, cleanedVal = False)
        self.assertEqual(len(d), 42)
        self.assertEqual(len(d['UT']), len(self.RC))
        self.assertEqual(d['EI'][0], None)
        self.assertIsInstance(d['CR'], list)

    def test_coCite(self):
        Gdefault = self.RC.coCiteNetwork(fullInfo = True)
        Gauths = self.RC.coCiteNetwork(nodeType = "author", dropAnon = False)
        GauthsNoExtra = self.RC.coCiteNetwork(nodeType = "author", nodeInfo = False)
        Gunwei = self.RC.coCiteNetwork(nodeType = 'original', weighted = False)
        Gjour = self.RC.coCiteNetwork(nodeType = "journal", dropNonJournals = True)
        Gyear = self.RC.coCiteNetwork(nodeType = "year", fullInfo = True)
        self.assertIsInstance(Gdefault, nx.classes.graph.Graph)
        self.assertLessEqual(len(Gdefault.edges()), len(Gunwei.edges()))
        self.assertLessEqual(len(Gdefault.nodes()), len(Gunwei.nodes()))
        self.assertEqual(len(GauthsNoExtra.edges()), len(Gauths.edges()))
        self.assertEqual(len(GauthsNoExtra.nodes()), len(Gauths.nodes()) - 1 )
        self.assertTrue('weight' in Gdefault.edges(data = True)[0][2])
        self.assertTrue('info' in Gdefault.nodes(data = True)[0][1])
        self.assertTrue('fullCite' in Gdefault.nodes(data = True)[0][1])
        self.assertFalse('weight' in Gunwei.edges(data = True)[0][2])
        self.assertEqual(len(Gdefault.nodes()), 518)
        self.assertEqual(len(Gdefault.edges()), 14775)
        self.assertEqual(len(Gauths.nodes()), 322)
        self.assertEqual(len(Gauths.edges()), 6996)
        self.assertEqual(len(Gyear.nodes()), 91)
        self.assertEqual(len(Gyear.edges()), 1962)
        self.assertEqual(len(Gjour.nodes()), 84)
        self.assertEqual(len(Gjour.edges()), 1215)
        self.assertTrue('info' in Gjour.nodes(data=True)[0][1])
        self.assertTrue('info' in Gyear.nodes(data=True)[0][1])
        self.assertTrue('fullCite' in Gyear.nodes(data = True)[0][1])

    def test_coAuth(self):
        Gdefault = self.RC.coAuthNetwork()
        self.assertIsInstance(Gdefault, nx.classes.graph.Graph)
        self.assertEqual(len(Gdefault.nodes()), 45)
        self.assertEqual(len(Gdefault.edges()), 46)

    def test_Cite(self):
        Gdefault = self.RC.citationNetwork(extraInfo = True)
        Ganon = self.RC.citationNetwork(dropAnon = False)
        Gauths = self.RC.citationNetwork(nodeType = "author")
        GauthsNoExtra = self.RC.citationNetwork(nodeType = "author", extraInfo = False)
        Gunwei = self.RC.citationNetwork(nodeType = 'original', weighted = False)
        Gjour = self.RC.citationNetwork(nodeType = "author", dropNonJournals = True, saveJournalNames = True)
        Gyear = self.RC.citationNetwork(nodeType = "year", saveJournalNames = True)
        self.assertIsInstance(Gdefault, nx.classes.digraph.DiGraph)
        self.assertLessEqual(len(Gdefault.edges()), len(Gunwei.edges()))
        self.assertLessEqual(len(Gdefault.nodes()), len(Gunwei.nodes()))
        self.assertEqual(len(GauthsNoExtra.edges()), len(Gauths.edges()))
        self.assertEqual(len(GauthsNoExtra.nodes()), len(Gauths.nodes()))
        self.assertTrue('weight' in Gdefault.edges(data = True)[0][2])
        self.assertTrue('info' in Gdefault.nodes(data = True)[0][1])
        self.assertFalse('weight' in Gunwei.edges(data = True)[0][2])
        self.assertGreater(len(Gdefault.nodes()), len(Gauths.nodes()))
        self.assertGreater(len(Ganon.nodes()), len(Gdefault.nodes()))
        self.assertEqual(len(Gdefault.nodes()), 535)
        self.assertEqual(len(Ganon.nodes()), 536)
        self.assertEqual(len(Gauths.nodes()), 325)
        self.assertEqual(len(Gdefault.edges()), 847)
        self.assertEqual(len(Ganon.edges()), 848)
        self.assertEqual(len(Gauths.edges()), 569)
        self.assertEqual(len(Gjour.edges()), 429)
        self.assertTrue('journal' in Gjour.nodes(data=True)[0][1])
        self.assertTrue('journal' in Gyear.nodes(data=True)[0][1])


    def test_oneMode(self):
        Gcr  = self.RC.oneModeNetwork('CR')
        Gcite = self.RC.oneModeNetwork('citations', nodeCount = False, edgeWeight = False)
        GcoCit = self.RC.coCiteNetwork()
        Gtit = self.RC.oneModeNetwork('title')
        self.assertEqual(len(Gcite.edges()), len(Gcr.edges()))
        self.assertEqual(len(Gcite.nodes()), len(Gcr.nodes()))
        self.assertAlmostEqual(len(Gcite.nodes()), len(GcoCit.nodes()), delta = 50)
        self.assertEqual(len(self.RC.oneModeNetwork('D2').nodes()), 0)
        self.assertEqual(len(Gtit.nodes()), 31)
        self.assertEqual(len(Gtit.edges()), 0)
        self.assertEqual(len(self.RC.oneModeNetwork('email').edges()), 3)
        self.assertEqual(len(self.RC.oneModeNetwork('UT').nodes()), len(self.RC) - 1)
        with self.assertRaises(TypeError):
            G = self.RC.oneModeNetwork('Not a Tag')
            del G

    def test_twoMode(self):
        self.RC.dropBadRecords()
        Gutti = self.RC.twoModeNetwork('UT', 'title', directed = True, recordType = False)
        Gafwc = self.RC.twoModeNetwork('AF', 'WC', nodeCount = False, edgeWeight = False)
        Gd2em = self.RC.twoModeNetwork('D2', 'email')
        Gemd2 = self.RC.twoModeNetwork('email', 'D2')
        self.assertIsInstance(Gutti, nx.classes.digraph.DiGraph)
        self.assertIsInstance(Gafwc, nx.classes.graph.Graph)
        self.assertEqual(Gutti.edges('WOS:A1979GV55600001')[0][1][:31], "EXPERIMENTS IN PHENOMENOLOGICAL")
        self.assertEqual(len(Gutti.nodes()), 2 * len(self.RC) - 1)
        with self.assertRaises(TypeError):
            G = self.RC.oneModeNetwork('Not a Tag', 'TI')
            del G
        self.assertTrue(nx.is_isomorphic(Gd2em, Gemd2))

    def test_nMode(self):
        G = self.RC.nModeNetwork(metaknowledge.tagToFull.keys())
        self.assertEqual(len(G.nodes()), 1186)
        self.assertEqual(len(G.edges()), 38592)

    def test_citeFilter(self):
        RCmin = self.RC.citeFilter('', reverse = True)
        RCmax = self.RC.citeFilter('')
        RCanon = self.RC.citeFilter('', 'anonymous')
        RC1970 = self.RC.citeFilter('1970', 'year')
        RCno1970 = self.RC.citeFilter(1970, 'year', reverse = True)
        RCMELLER = self.RC.citeFilter('meller', 'author')
        self.assertEqual(len(RCmin), 0)
        self.assertEqual(len(RCmax), len(self.RC))
        self.assertEqual(len(RCanon), 1)
        self.assertEqual(len(RC1970), 15)
        self.assertEqual(len(RC1970) + len(RCno1970), len(self.RC))
        self.assertEqual(len(RCMELLER), 1)
        RCnocite = metaknowledge.RecordCollection('metaknowledge/tests/OnePaperNoCites.isi')
        self.assertEqual(len(RCnocite.citeFilter('')), 0)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
from .recordCollection import RecordCollection

import sys
import io

def btest(quite = False):
    if quite:
        stdOut = sys.stdout
        sys.stdout = io.StringIO()
    currentPath = '.' #os.path.dirname(os.path.realpath(__file__))
    R1 = RecordCollection(currentPath + "/metaknowledge/tests/testFile.isi")
    R2 = RecordCollection(currentPath + "/metaknowledge/tests/ManyAuthors.isi")
    print(R1)
    print("Network")
    print(R1.coAuthNetwork().edges)
    R3 = R1 + R2
    print(R3.coAuthNetwork().edges())
    print(R3.pop().year)
    print(R3.pop())
    print(dir(R3))
    r1 = R3.pop()
    print(r1)
    for r in R3:
        if r.wosString:
            print(r.wosString, end = ': \t')
            print(r.month, end = ',\t')
            print(r.year)
        else:
            print(r)
    print(R3)
    print(R2)
    print(R1)
    print(R2 & R1)
    print(R2 + R2)
    print(R2 ^ R1)
    print(R3 & R1)
    print(repr(R1))
    print(repr(R3 + R2 - R3 & R1))

    for r in R1:
        if r.bad:
            print(r.error)
        else:
            print(r)
            print(r.authorsFull)
            print(r.year)
            print(r.month)

            print(r.title)
            print(r.month)
            print(r.wosString)
            print(r.DOI)
            print(r.citations)
    print(R3.pop())
    if quite:
        sys.stdout = stdOut



import metaknowledge

import networkx as nx
import csv
import os
import sys
import time
import math

def read_graph(edgeList, nodeList = None, directed = False, idKey = 'ID', eSource = 'From', eDest = 'To'):
    """
    Reads the files given by edgeList and if given nodeList and produces a networkx graph
    This is designed only for the files produced by metaknowledge and is meant to be the reverse of write_graph()

    nodeList must be given if any of the attributes of the node are needed
    directed controls if the resultant graph is directional eSource and eDest control the direction
    idKey, eSource and  eDest are the labels for the edge's id, source and destination respectively, they must match headers in the file or a keyError exception will be thrown

    igraph Style
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Starting to reading graphs")
    else:
        PBar = None
    if directed:
        grph = nx.DiGraph()
    else:
        grph = nx.Graph()
    if nodeList:
        if PBar:
            PBar.updateVal(0, "Reading " + nodeList)
        f = open(nodeList)
        nFile = csv.DictReader(f)
        for line in nFile:
            vals = line
            ndID = vals[idKey]
            del vals[idKey]
            if len(vals) > 0:
                grph.add_node(ndID, attr_dict=vals)
            else:
                grph.add_node(ndID)
        f.close()
    if PBar:
        PBar.updateVal(.5, "Reading " + edgeList)
    f = open(edgeList)
    eFile = csv.DictReader(f)
    for line in eFile:
        vals = line
        eFrom = vals[eSource]
        eTo = vals[eDest]
        del vals[eSource]
        del vals[eDest]
        if len(vals) > 0:
            grph.add_edge(eFrom, eTo, attr_dict = vals)
        else:
            grph.add_edge(eFrom, eTo)
    if PBar:
        PBar.finish(str(len(grph.nodes())) + " nodes and " + str(len(grph.edges())) + " edges found")
    f.close()
    return grph

def write_graph(grph, name, edgeInfo = True, typing = True, suffix = 'csv', overwrite = False):
    """
    Writes both the edge list and the node attribute list of grph.
    The output files start with name, the file type[edgeList, nodeAttributes] then if typing is True the type of graph (directed or undirected) then the suffix, it appears as follows:
        name_fileType_Graphtype.suffix
    If edgeInfo is true the extra information about the edges will be included in their list, i.e. their weight will be given
    If overwrite is False write_graph will throw an exception if either of the files it is attempting to write exist
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Writing the graph to two files starting with: " + name)
    else:
        PBar = None
    if typing:
        if isinstance(grph, nx.classes.digraph.DiGraph) or isinstance(grph, nx.classes.multidigraph.MultiDiGraph):
            grphType = "_directed"
        else:
            grphType = "_undirected"
    else:
        grphType = ''
    nameCompts = os.path.split(os.path.expanduser(os.path.normpath(name)))
    if nameCompts[0] == '' and nameCompts[1] == '':
        edgeListName = "edgeList"+ grphType + '.' + suffix
        nodesAtrName = "nodeAttributes"+ grphType + '.' + suffix
    elif nameCompts[0] == '':
        edgeListName = nameCompts[1] + "_edgeList"+ grphType + '.' + suffix
        nodesAtrName = nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix
    elif nameCompts[1] == '':
        edgeListName = os.path.join(nameCompts[0], "edgeList"+ grphType + '.' + suffix)
        nodesAtrName = os.path.join(nameCompts[0], "nodeAttributes"+ grphType + '.' + suffix)
    else:
        edgeListName = os.path.join(nameCompts[0], nameCompts[1] + "_edgeList"+ grphType + '.' + suffix)
        nodesAtrName = os.path.join(nameCompts[0], nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix)
    if not overwrite:
        if os.path.isfile(edgeListName):
            raise OSError(edgeListName+ " already exists")
        if os.path.isfile(nodesAtrName):
            raise OSError(nodesAtrName + " already exists")
    write_edgeList(grph, edgeListName, extraInfo = edgeInfo, progBar = PBar)
    if PBar:
        PBar.jumpUp()
    write_nodeAttributeFile(grph, nodesAtrName, progBar = PBar)
    if PBar:
        PBar.finish(str(len(grph.nodes())) + " nodes and " + str(len(grph.edges())) + " edges written to file")

def write_edgeList(grph, name, extraInfo = True, progBar = None):
    """
    writes an edge list of grph with filename name, if extraInfo is true the additional information about the edges, e.g. weight, will be written.
    All edges must have the same tags
    """
    if progBar:
        count = 0
        eMax = len(grph.edges(data = True))
        if isinstance(progBar, _ProgressBar):
            progBar.updateVal(0, "Writing edge list " + name)
        else:
            progBar = _ProgressBar(0, "Writing edge list " + name)
    if len(grph.edges(data = True)) < 1:
        outFile = open(name, 'w')
        outFile.write('"From","To"\n')
        outFile.close()
        if progBar:
            progBar.updateVal(1, "Done edge list " + name + ", 0 edges written.")
    else:
        if extraInfo:
            csvDict = ['From'] +  ['To'] + list(grph.edges_iter(data = True).__next__()[2].keys())
        else:
            csvDict = ['From'] +  ['To']
        f = open(name, 'w')
        outFile = csv.DictWriter(f, csvDict, delimiter = ',', quotechar = '"', quoting=csv.QUOTE_ALL)
        outFile.writeheader()
        if extraInfo:
            for e in grph.edges_iter(data = True):
                if progBar:
                    count += 1
                    if count % 10000 == 0:
                        progBar.updateVal(count / eMax)
                eDict = e[2].copy()
                eDict['From'] = e[0]
                eDict['To'] = e[1]
                try:
                    outFile.writerow(eDict)
                except ValueError:
                    raise ValueError("Some edges in " + str(grph) + " do not have the same attributes")
        else:
            for e in grph.edges_iter():
                if progBar:
                    count += 1
                    if count % 100 == 0:
                        progBar.updateVal(count / eMax)
                eDict['From'] = e[0]
                eDict['To'] = e[1]
                outFile.writerow(eDict)
        if progBar:
            progBar.finish("Done edge list " + name + ", " + str(count) + " edges written.")
        f.close()

def write_nodeAttributeFile(grph, name, progBar = None):
    """
    writes a node attribute list of grph with filename name, the first column is the node's ID then all after it are its associated information.
    All nodes must have the same tags.
    """
    if progBar:
        count = 0
        nMax = len(grph.nodes())
        if isinstance(progBar, _ProgressBar):
            progBar.updateVal(0, "Writing edgelist " + name)
        else:
            progBar = _ProgressBar(0, "Writing edgelist " + name)
    if len(grph.nodes(data = True)) < 1:
        outFile = open(name, 'w')
        outFile.write('ID\n')
        outFile.close()
        if progBar:
            progBar.updateVal(1, "Done node attribute list: " + name + ", 0 nodes written.")
    else:
        csvDict = ['ID'] + list(grph.nodes_iter(data = True).__next__()[1].keys())
        f = open(name, 'w')
        outFile = csv.DictWriter(f, csvDict, delimiter = ',', quotechar = '"', quoting=csv.QUOTE_ALL)
        outFile.writeheader()
        for n in grph.nodes_iter(data = True):
            if progBar:
                count += 1
                if count % 50 == 0:
                    progBar.updateVal(count / nMax)
            nDict = n[1].copy()
            nDict['ID'] = n[0]
            try:
                outFile.writerow(nDict)
            except ValueError:
                raise ValueError("Some nodes in " + str(grph) + " do not have the same attributes")
        if progBar:
            progBar.updateVal(1, "Done node attribute list: " + name + ", " + str(count) + " nodes written.")
        f.close()

class _ProgressBar(object):
    difTermAndBar = 8 #the number of characters difference between the bar's length and the terminal's width
    timeLength = 6 # width of elapse time display
    def __init__(self, initPer, initString = ' ', output = sys.stdout):
        self.finished = False
        self.per = initPer
        self.out = output
        self.sTime = time.time()
        try:
            self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
            if self.barMaxLength < 0:
                self.barMaxLength = 0
        except OSError:
            self.barMaxLength = 80 - self.difTermAndBar
        self.dString = self.prepString(initString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength)
        self.out.write('[' + ' ' * self.barMaxLength + ']' + '{:.1%}'.format(self.per).rjust(6, ' ') + '\n')
        self.out.write(self.dString + '\033[F')
        self.out.flush()

    def __del__(self):
        if not self.finished:
            self.out.write('\n\n')
            self.out.flush()

    def updateVal(self, inputPer, inputString = None):
        try:
            self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
            if self.barMaxLength < 0:
                self.barMaxLength = 0
        except OSError:
            self.barMaxLength = 80 - self.difTermAndBar
        self.out.write('\r')
        self.per = inputPer
        percentString = '{:.1%}'.format(self.per).rjust(6, ' ')
        barLength = int(self.per * self.barMaxLength)
        if inputString:
            self.dString = self.prepString(inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength)
            if barLength >= self.barMaxLength:
                self.out.write('[' + '=' * barLength + ']' + percentString)
                self.out.write('\n' + self.dString + '\033[F')
            else:
                self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString)
                self.out.write('\n' + self.dString + '\033[F')
        else:
            if barLength >= self.barMaxLength:
                self.out.write('[' + '=' * barLength + ']' + percentString + '\r')
            else:
                self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString + '\r')
        self.out.flush()

    def finish(self, inputString):
        self.finished = True
        inputString = str(inputString)
        try:
            self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
            if self.barMaxLength < 0:
                self.barMaxLength = 0
        except OSError:
            self.barMaxLength = 80 - self.difTermAndBar
        self.out.write('\n' + ' ' * (self.barMaxLength + self.difTermAndBar) + '\033[F')
        if len(inputString) < self.barMaxLength + self.difTermAndBar - self.timeLength:
            tString = self.prepTime(time.time() - self.sTime, self.barMaxLength + self.difTermAndBar - len(inputString) - 1)
            self.out.write(inputString + ' ' + tString)
        else:
            self.out.write(self.prepString(inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength))
        self.out.write('\n')
        self.out.flush()

    def jumpUp(self):
        self.out.write('\033[F')
        self.out.flush()

    @staticmethod
    def prepString(s, maxLength):
        maxLength = maxLength - 1
        sString = str(s)
        if len(sString) <= maxLength:
            return sString.ljust(maxLength, ' ') + ' '
        else:
            if maxLength % 2 == 0:
                return sString[:int(maxLength/2 - 3)] + '...' + sString[int(-maxLength/2):] + ' '
            else:
                return sString[:int(maxLength/2 - 2)] + '...' + sString[int(-maxLength/2):] + ' '
    @staticmethod
    def prepTime(t, maxLength):
        try:
            if math.log10(t) + 3.01 > maxLength:
                return "{1:{0}.0E}s".format(maxLength - 1 ,t)
            else:
                return "{1:{0}.1f}s".format(maxLength - 1,t)
        except ValueError:
            return "{1:{0}.1f}s".format(maxLength - 1,t)


def getWeight(grph, nd1, nd2, weightString = "weight", returnType = int):
    """
    A way of getting the weight of an edge with or without weight as a parameter
    returns a the value of the weight parameter converted to returnType if it is given or 1 (also converted) if not
    """
    if not weightString:
        return returnType(1)
    else:
        return returnType(grph.edge[nd1][nd2][weightString])

def getNodeDegrees(grph, weightString = "weight", strictMode = False,  returnType = int, edgeType = 'bi'):
    """
    Retunrs a dictionary of nodes to their degrees, the degree is determined by adding the weight of edge with the weight being the string weightString that gives the name of the attribute of each edge containng thier weight. The Weights are then converted to the type returnType. If weightString is give as False instead each edge is counted as 1.

    edgeType, takes in one of three strings: 'bi', 'in', 'out'. 'bi' means both nodes on the edge count it, 'out' mans only the one the edge comes form counts it and 'in' means only the node the edge goes to counts it. 'bi' is the default. Use only on directional graphs as otherwise the selected nodes is random.
    """
    ndsDict = {}
    for nd in grph.nodes_iter():
        ndsDict[nd] = returnType(0)
    for e in grph.edges_iter(data = True):
        if weightString:
            try:
                edgVal = returnType(e[2][weightString])
            except KeyError:
                if strictMode:
                    raise KeyError("The edge from " + str(e[0]) + " to " + str(e[1]) + " does not have the attribute: '" + str(weightString) + "'")
                else:
                    edgVal = returnType(1)
        else:
            edgVal = returnType(1)
        if edgeType == 'bi':
            ndsDict[e[0]] += edgVal
            ndsDict[e[1]] += edgVal
        elif edgeType == 'in':
            ndsDict[e[1]] += edgVal
        elif edgeType == 'out':
            ndsDict[e[0]] += edgVal
        else:
            raise ValueError("edgeType must be 'bi', 'in', or 'out'")
    return ndsDict

def getDegreeDistribution(grph, weightParameter = "weight", strictWeightNames = False,  weightType = int, directionalType = 'bi'):
    if weightType != int:
        raise ValueError("Unsupported type for weights, only ints are supported")
    ndsDict = getNodeDegrees(grph, weightString = weightParameter, strictMode = strictWeightNames, returnType = weightType, edgeType = directionalType)
    distVec = [0] * (max(ndsDict.values()) + 1)
    for v in ndsDict.values():
        distVec[v] += 1
    return distVec

def drop_edges(grph, minWeight = - float('inf'), maxWeight = float('inf'), parameterName = 'weight', ignoreUnweighted = False, dropSelfLoops = False):
    """
    Returns a graph with edges whose weight is within the inclusive bounds of minWeight and maxWeight, i.e minWeight <= edges weight <= maxWeight, will throw a Keyerror if the graph is unweighted

    minWeight and maxWeight default to negative and positive infinity respectively so without specifying either the output should be the input

    parameterName is key to weight field in the edge's dictionary, default is weight as that is almost always correct

    ignoreUnweighted can be set False to suppress the KeyError and make unweighted edges be ignored
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping edges")
        count = 0
        total = len(grph.edges())
    else:
        PBar = None
    tmpGrph = grph.copy()
    if minWeight != - float('inf') or maxWeight != float('inf'):
        for e in grph.edges_iter(data = True):
            try:
                val = e[2][parameterName]
            except KeyError:
                if not ignoreUnweighted:
                    raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                else:
                    pass
            else:
                if PBar:
                    count += 1
                    if count % 100000 == 0:
                        PBar.updateVal(count/ total, str(count) + " edges analysed and " + str(total -len(tmpGrph.edges())) + " edges dropped")
                if val > maxWeight or  val < minWeight:
                    tmpGrph.remove_edge(e[0], e[1])
    if dropSelfLoops:
        slps = tmpGrph.selfloop_edges()
        if PBar:
            PBar.updateVal(1, "Dropping self {} loops".format(len(slps)))
        for e in slps:
            tmpGrph.remove_edge(e[0], e[1])
    if PBar:
        PBar.finish(str(total - len(tmpGrph.edges())) + " edges out of " + str(total) + " dropped, " + str(len(tmpGrph.edges())) + " returned")
    return tmpGrph

def drop_nodesByDegree(grph, minDegree = -float('inf'), maxDegree = float('inf'), useWeight = True, parameterName = 'weight', ignoreUnweighted = True):
    """
    Returns a graph whose nodes have a degree that is within inclusive bounds of minDegree and maxDegree, i.e minDegree <= degree <= maxDegree. Degree can be determined in two ways by default it is the total number of edges touching a node, alternative if useWeight is True it is the sum of the weight of all the edges touching a node.

    minDegree and maxDegree default to negative and positive infinity respectively so without specifying either the output should be the input

    useWeight can be set True to use an alternative method for calculating degree, the total weight of all edges

    parameterName is key to weight field in the edge's dictionary, default is weight as that is almost always correct, only used if useWeight is True

    ignoreUnweighted can be set False to suppress the KeyError and make unweighted edges be not counted, only used if useWeight is True
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping nodes by degree")
        count = 0
        total = len(grph.nodes())
    else:
        PBar = None
    goodNodes = []
    for n in grph.nodes_iter():
        if PBar:
            count += 1
            if count % 10000 == 0:
                PBar.updateVal(count/ total, str(count) + " nodes analysed and " + str(total - len(goodNodes)) + " nodes dropped")
        val = 0
        if useWeight:
            for e in grph.edges(n, data = True):
                try:
                    val += e[2][parameterName]
                except KeyError:
                    if not ignoreUnweighted:
                        raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                    else:
                        val += 1
        else:
            val = len(grph.edges(n))
        if val <= maxDegree and val >= minDegree:
            goodNodes.append(n)
    if PBar:
        PBar.finish(str(total - len(goodNodes)) + " nodes out of " + str(total) + " dropped, " + str(len(goodNodes)) + " returned")
    return grph.subgraph(goodNodes)

def drop_nodesByCount(grph, minCount = -float('inf'), maxCount = float('inf'), parameterName = 'count', ignoreMissing = False):
    """
    Returns a graph whose nodes have a occurrence count that is within inclusive bounds of minCount and maxCount, i.e minCount <= count <= maxCount. Occurrence count is determined by reading the variable associated with the node named parameterName.

    minCount and maxCount default to negative and positive infinity respectively so without specifying either the output should be the input


    parameterName is key to count field in the node's dictionary, default is count as that is often correct

    ignoreMissing can be set False to suppress the KeyError and make nodes missing counts be dropped instead of throwing errors
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping nodes by count")
        count = 0
        total = len(grph.nodes())
    else:
        PBar = None
    goodNodes = []
    for n in grph.nodes_iter(data = True):
        if PBar:
            count += 1
            if count % 10000 == 0:
                PBar.updateVal(count/ total, str(count) + "nodes analysed and " + str(total - len(goodNodes)) + " nodes dropped")

        try:
            val = n[1][parameterName]
        except KeyError:
            if not ignoreMissing:
                raise KeyError("One or more nodes do not have counts or " + str(parameterName), " is not the name of the count parameter")
            else:
                pass
        else:
            if val <= maxCount and val >= minCount:
                goodNodes.append(n[0])
    if PBar:
        PBar.finish(str(total - len(goodNodes)) + " nodes out of " + str(total) + " dropped, " + str(len(goodNodes)) + " returned")
    return grph.subgraph(goodNodes)

def graphStats(G, stats = ('nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity'), makeString = True):
    for sts in stats:
        if sts not in ['nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity']:
            raise RuntimeError('"{}" is not a valid stat.'.format(sts))
    if makeString:
        stsData = []
    else:
        stsData = {}
    if 'nodes' in stats:
        if makeString:
            stsData.append("{:G} nodes".format(len(G.nodes())))
        else:
            stsData['nodes'] = len(G.nodes())
    if 'edges' in stats:
        if makeString:
            stsData.append("{:G} edges".format(len(G.edges())))
        else:
            stsData['edges'] = len(G.edges())
    if 'isolates' in stats:
        if makeString:
            stsData.append("{:G} isolates".format(len(nx.isolates(G))))
        else:
            stsData['isolates'] = len(nx.isolates(G))
    if 'loops' in stats:
        if makeString:
            stsData.append("{:G} self loops".format(len(G.selfloop_edges())))
        else:
            stsData['loops'] = len(G.selfloop_edges())
    if 'density' in stats:
        if makeString:
            stsData.append("a density of {:G}".format(nx.density(G)))
        else:
            stsData['density'] = nx.density(G)
    if 'transitivity' in stats:
        if makeString:
            stsData.append("a transitivity of {:G}".format(nx.transitivity(G)))
        else:
            stsData['transitivity'] = nx.transitivity(G)
    if makeString:
        retString = "The graph has "
        if len(stsData) < 1:
            return retString
        elif len(stsData) == 1:
            return retString + stsData[0]
        else:
            return retString + ', '.join(stsData[:-1]) + ' and ' + stsData[-1]
    else:
        retLst = []
        for sts in stats:
            retLst.append(stsData[sts])
        return tuple(retLst)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import metaknowledge
from .record import Record, BadISIFile
from .graphHelpers import _ProgressBar
from .constants import tagsAndNames, tagToFull, fullToTag
from .citation import filterNonJournals

import itertools
import os.path
import csv

import networkx as nx

class RecordCollection(object):
    """
    A way of containing a large number of Record objects, it provides ways of creating them from an isi file, string, list of records or directory containing isi files. The Records are containing within a set and as such many of the set operations are defined, pop, union, in ... also records are hashed with their WOS string so no duplication can occur.
    The comparison operators <, <=, >, >= are based strictly on the number of Records within the collection, while equality looks for an exact match on the Records

    When being created if there are issues the Record collection will be declared bad (self.bad = True) it will then mostly return nothing or False. The error attribute contains the exception that occurred.

    They also possess a name accessed with repr(), this is used to auto generate the names of files and can be set at creation, note though that any operations that modify the RecordCollection's contents will update the name to include what occurred, read __repr__'s doc string for more information

    inCollection is the object containing the information about the Records to be constructed it can be an isi file, string, list of records or directory containing isi files

    name sets the name of the of the record if left blank name will be generated based on the object that created the Recordcollection

    extension controls the extension that __init__ looks for when reading a directory, set it to the extension on the isi files you wish to load, if left blank all files will be tried and any that are not isi files will be silently skipped

    # \_\_Init\_\_

    RecordCollections are made from either a single file or directory supplied as _inCollection_.

    # Parameters

    _inCollection_ : `optional [str] or None`

    > the name of the source of WOS records. It can be skipped to produce an empty collection.

    > If a file is provided. First it is checked to see if it is a WOS file (the header is checked). Then records are read from it one by one until the 'EF' string is found indicating the end of the file.

    > If a directory is provided. First each file in the directory is checked for the correct header and all those that do are then read like indivual files. The records are then collected into a single set in the RecordCollection.

    _name_ : `optional [str]`

    > The name of the RecordCollection, defaults to empty string. If left empty the name of the Record collection is set to the name of the file or directory used to create the collection. If provided the name id set to _name_

    _extension_ : `optional [str]`

    > The extension to search for when reading a directoy for files. _extension_ is the suffix searched for when a direcorty is read for files, by default it is empty so all files are read.
    """

    def __init__(self, inCollection = None, name = '', extension = ''):
        self.bad = False
        self._repr = name
        if not inCollection:
            if not name:
                self._repr = "empty"
            self._Records = set()
        elif isinstance(inCollection, str):
            if os.path.isfile(inCollection):
                try:
                    if not inCollection.endswith(extension):
                        raise TypeError("extension of input file does not match requested extension")
                    self._repr = os.path.splitext(os.path.split(inCollection)[1])[0]
                    self._Records = set(isiParser(inCollection))
                except BadISIFile as w:
                    self.bad = True
                    self.error = w
            elif os.path.isdir(inCollection):
                if metaknowledge.VERBOSE_MODE:
                    PBar = _ProgressBar(0, "Reading files from " + str(inCollection))
                    count = 0
                else:
                    PBar = None
                if extension and not name:
                    if extension[0] == '.':
                        self._repr = extension[1:] + "-files-from-" + os.path.dirname(inCollection)
                    else:
                        self._repr = extension + "-files-from-" + inCollection
                elif not name:
                    self._repr = "files-from-" + inCollection
                self._Records = set()
                flist = []
                for f in os.listdir(inCollection):
                    fullF = os.path.join(os.path.abspath(inCollection), f)
                    if fullF.endswith(extension) and os.path.isfile(fullF):
                        flist.append(fullF)
                for file in flist:
                    if PBar:
                        count += 1
                        PBar.updateVal(count / len(flist), "Reading records from: " + file)
                    else:
                        PBar = None
                    try:
                        self._Records |= set(isiParser(file))
                    except BadISIFile:
                        if extension != '':
                            raise
                        else:
                            pass
                    except UnicodeDecodeError:
                        pass
                if PBar:
                    PBar.finish("Done reading records from: " + str(inCollection))
            else:
                raise TypeError("inCollection is not a directory or a file")
        elif isinstance(inCollection, list):
            self._Records = set(inCollection)
        elif isinstance(inCollection, set):
            self._Records = inCollection
        else:
            raise TypeError

    def __add__(self, other):
        """
        returns the union of the two RecordCollections
        """
        if self.bad and other.bad:
            return RecordCollection(set(), '[BAD ' + repr(self) + ']'+ '_plus_' + '[BAD ' +  repr(other) + ']')
        if self.bad:
            return RecordCollection(other._Records, '[BAD ' + repr(self) + ']' + '_plus_' + repr(other))
        elif other.bad:
            return RecordCollection(self._Records,  repr(self) + '[BAD ' + repr(other) + ']')
        else:
            return RecordCollection(self._Records | other._Records, repr(self) + '_plus_' + repr(other))

    def __and__(self, other):
        """
        returns the intersection of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records & other._Records, repr(self) + '_and_' + repr(other))

    def __sub__(self, other):
        """
        returns the difference of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records - other._Records, repr(self) + '_diff_' + repr(other))

    def __xor__(self, other):
        """
        returns the symmetric difference of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records ^ other._Records, repr(self) + '_symdiff_' + repr(other))

    def __str__(self):
        """
        Returns a string giving the the number Records in a sentence:
        "Collection of # records"
        """
        return "Collection of " + str(len(self._Records)) + " records"

    def __repr__(self):
        """
        The name of the RecordCollection, this used to identify the Collection when it is written as a file by writeFile() or in some of CL scripts
        It is updated when some modification of the RecordCollection occurs i.e.
        >>> RC = metaknowledge.RecordCollection('.', extension = 'isi')
        >>> repr(RC)
        isi-files-from-.
        >>> R = RC.pop()
        >>> repr(RC)
        pop-isi-files-from-.
        """
        return self._repr

    def __lt__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) < len(other)
    def __le__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) <= len(other)
    def __gt__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) > len(other)
    def __ge__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) >= len(other)

    def __eq__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return self._Records == other._Records

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        """
        returns the number or Records
        """
        return len(self._Records)

    def __iter__(self):
        """
        iterates over the Records
        """
        for R in self._Records:
            yield R

    def __contains__(self, item):
        """
        Returns True if the item is a WOS string of one of the Records containing or if item is a Record checks if the record is in the RecordCollection, otherwise returns False
        """
        if isinstance(item, str):
            for R in self:
                if R.UT == item:
                    return True
            return False
        elif isinstance(item, Record):
            return item in self._Records
        else:
            return False


    def pop(self):
        """
        Returns a random Record from the recordCollection, the Record is deleted from the collection, use peak for nondestructive access
        """
        if len(self._Records) > 0:
            self._repr = "Pop-" + self._repr
            return self._Records.pop()
        else:
            return None

    def peak(self):
        """
        Returns a random Record from the recordCollection, the Record is kept in the collection, use pop for faster destructive access
        """
        if len(self._Records) > 0:
            return self._Records.__iter__().__next__()
        else:
            return None

    def dropWOS(self, wosNum):
        """Removes the Record with WOS number (ID number) _wosNum_

        # Parameters

        _wosNum_ : `str`

        > _wosNum_ is the WOS number of the Record to be dropped. _wosNum_ must begin with 'WOS:' or a valueError is raise.
        """
        if wosNum[:4] != 'WOS:':
            raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        for R in self._Records:
            if R.wosString == wosNum:
                self._Records.remove(R)
                break

    def getWOS(self, wosNum, drop = False):
        """Gets the Record from the collection by its WOS number.

        # Parameters

        _wosNum_ : `str`

        > _wosNum_ is the WOS number of the Record to be extracted. _wosNum_ must begin with 'WOS:' or a valueError is raise.

        _drop_ : `optional [bool]`

        > Default `False`. If `True` the Record is dropped from the collection after being extract, i.e. if `False` [getWOS()](#RecordCollection.getWOS) acts like [peak()](#RecordCollection.peak), if `True` it acts like [pop()](#RecordCollection.pop)

        # Returns

        `metaknowledge.Record`

        > The Record whose WOS number is _wosNum_
        """
        try:
            if wosNum[:4] != 'WOS:':
                raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        except (TypeError, IndexError):
            raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        for R in self:
            if R.wosString == wosNum:
                if drop:
                    self._Records.remove(R)
                return R
        return None

    def getBadRecords(self):
        """
        returns RecordCollection containing all the Record which have their bad flag set to True, i.e. all those removed by dropBadRecords()
        """
        badRecords = set()
        for R in self._Records:
            if R.bad:
                badRecords.add(R)
        return RecordCollection(badRecords, repr(self) + '_badRecords')

    def dropBadRecords(self):
        """
        Removes all Records with bad attributes == True from the collection
        """
        self._Records = {r for r in self._Records if not r.bad}
        self._repr = repr(self) + '_badRecordsDropped'

    def dropNonJournals(self, ptVal = 'J', dropBad = True, invert = False):
        """Drops the non journal type Records from the collection

        # Parameters

        _ptVal_ : `optional [str]`

        > The value of the PT tag to be kept, default is 'J' the journal tag

        _dropBad_ : `optional [bool]`

        > Determines if bad Records will be dropped as well, default `True`

        _invert_ : `optional [bool]`

        > Set `True` to drop journals (or the PT tag given by _ptVal) instead of keeping them. Note, it still drops bad Records if _dropBad_ is `True`, default `False`
        """
        if dropBad:
            self.dropBadRecords()
        if invert:
            self._Records = {r for r in self._Records if r.PT != ptVal.upper()}
            self._repr = repr(self) + '_PT-{}-Dropped'.format(ptVal)
        else:
            self._Records = {r for r in self._Records if r.PT == ptVal.upper()}
            self._repr = repr(self) + '_PT-{}-Only'.format(ptVal)

    def writeFile(self, fname = None):
        """
        Writes the RecordCollection to a file, the written file is identical to those download from WOS. The order of Records written is random.

        fname set the name of the file, if blank the RecordCollection's name's first 200 characters are use with the suffix .isi
        """
        if fname:
            f = open(fname, mode = 'w', encoding = 'utf-8')
        else:
            f = open(repr(self)[:200] + '.isi', mode = 'w', encoding = 'utf-8')
        f.write("\ufeffFN Thomson Reuters Web of Science\u2122\n")
        f.write("VR 1.0\n")
        for R in self._Records:
            R.writeRecord(f)
            f.write('\n')
        f.write('EF')
        f.close()

    def writeCSV(self, fname = None, onlyTheseTags = None, longNames = False, firstTags = ['UT', 'PT', 'TI', 'AF', 'CR'], csvDelimiter = ',', csvQuote = '"', listDelimiter = '|'):
        """Writes all the Records from the collection into a csv file with each row a record and each column a tag

        fname is the name of the file to write to, if none is given it uses the Collections name suffixed by .csv

        onlyTheseTags lets you specify which tags to use, if not given then all tags in the records are given.
        If you want to use all known tags the use onlyTheseTags = metaknowledge.knownTagsList

        longNames if set to True will convert the tags to their longer names, otherwise the short 2 character ones will be used

        firstTags is the column's tags, it is set by default to ['UT', 'PT', 'TI', 'AF', 'CR'] so those will be written first if given by onlyTheseTags.
        Note if tags are in firstTags but not in onlyTheseTags, onlyTheseTags will override onlyTheseTags

        csvDelimiter is the delimiter used for the cells of the csv file, default is the comma (,)

        csvQuote is  the quote character used for the csv, default is the double quote (")

        listDelimiter is the delimiter used between values of the same cell if the tag for that record has multiple outputs, default is the pipe (|)
        """
        for i in range(len(firstTags)):
            if firstTags[i] in fullToTag:
                firstTags[i] = fullToTag[firstTags[i]]
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTag:
                    onlyTheseTags[i] = fullToTag[onlyTheseTags[i]]
            retrievedFields = [t for t in firstTags if t in onlyTheseTags] + [t for t in onlyTheseTags if t not in firstTags]
        else:
            retrievedFields = firstTags
            for R in self:
                tagsLst = [t for t in R.activeTags() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFull[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        if fname:
            f = open(fname, mode = 'w', encoding = 'utf-8')
        else:
            f = open(repr(self)[:200] + '.csv', mode = 'w', encoding = 'utf-8')
        csvWriter = csv.DictWriter(f, retrievedFields, delimiter = csvDelimiter, quotechar = csvQuote, quoting=csv.QUOTE_ALL)
        csvWriter.writeheader()
        for R in self:
            recDict = R.getTagsDict(retrievedFields)
            for k in recDict.keys():
                value = recDict[k]
                if hasattr(value, '__iter__'):
                    recDict[k] = listDelimiter.join([str(v) for v in value])
                elif recDict[k] == None:
                    recDict[k] = ''
                else:
                    recDict[k] = str(value)
            csvWriter.writerow(recDict)
        f.close()


    def makeDict(self, onlyTheseTags = None, longNames = False, cleanedVal = True):
        """Returns a dict with each key a tag and the values being lists of the values for each of the Records in the collection, `None` is given when there is no value and they are in the same order across each tag.

        When used in pandas: `pandas.DataFrame(RC.makeDict())` returns a data frame with each column a tag and each row a Record.

        # Parameters

        See writeCSV()
        """
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTag:
                    onlyTheseTags[i] = fullToTag[onlyTheseTags[i]]
            retrievedFields = onlyTheseTags
        else:
            retrievedFields = []
            for R in self:
                tagsLst = [t for t in R.activeTags() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFull[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        retDict = {k : [] for k in retrievedFields}
        for R in self:
            for k, v in R.getTagsDict(retrievedFields, cleaned = cleanedVal).items():
                retDict[k].append(v)
        return retDict

    def coAuthNetwork(self):
        """Creates a coauthorship network for the RecordCollection.

        # Returns

        `Networkx Graph`

        > A networkx graph with author names as nodes and collaborations as edges.
        """


        grph = nx.Graph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a co-authorship network")
            count = 0
        else:
            PBar = None
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count/ len(self), "Analyzing: " + str(R))
            if R.authorsFull and len(R.authorsFull) > 1:
                for i, auth1 in enumerate(R.authorsFull):
                    if auth1 not in grph:
                        grph.add_node(auth1, count = 1)
                    else:
                        grph.node[auth1]['count'] += 1
                    for auth2 in R.authorsFull[i + 1:]:
                        if auth2 not in grph:
                            grph.add_node(auth2, count = 1)
                        else:
                            grph.node[auth2]['count'] += 1
                        if grph.has_edge(auth1, auth2):
                            grph.edge[auth1][auth2]['weight'] += 1
                        else:
                            grph.add_edge(auth1, auth2, weight = 1)
            if R.authorsFull and len(R.authorsFull) == 1:
                auth1 = R.authorsFull[0]
                if auth1 not in grph:
                    grph.add_node(auth1, count = 1)
                else:
                    grph.node[auth1]['count'] += 1
        if PBar:
            PBar.finish("Done making a co-authorship network")
        return grph

    def coCiteNetwork(self, dropAnon = True, nodeType = "full", nodeInfo = True, fullInfo = False, weighted = True, dropNonJournals = False):
        """Creates a co-citation network for the RecordCollection.

        # Parameters

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _nodeType : `optional [str]`

        > default `Full`, can alos be "original", "author", "journal" or "year"

        _extraInfo_ : `optional [bool]`

        > default `True`, wether the original citation string is added to the node as an extra value, if `True` it is

        _weighted_ : `optional [bool]`

        > default `True`, wether the edges are weighted. If `True` the edges are weighted by the number of occurrences of the co-citation.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, wether to drop ciations of non-journals

        _saveJournalNames_ : `optional [bool]`

        > default `False`, wether to save the full name of the journals of citations. Citations missing a journal will have the string "None", use with _dropNonJournals_ to avoid this

        # Returns

        `Networkx Graph`

        > A networkx graph with hashes as ID and co-citation as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise ValueError("{} is not an allowed nodeType.".format(nodeType))
        tmpgrph = nx.Graph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a co-citation network")
            count = 0
        else:
            PBar = None
        if nodeType != allowedTypes[0]:
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count / len(self), "Analyzing: " + str(R))
                Cites = R.citations
                if Cites:
                    if dropNonJournals:
                        Cites = filterNonJournals(Cites)
                    if dropAnon:
                        filteredCites = [c for c in Cites if not c.isAnonymous() and hasattr(c, nodeType)]
                    else:
                        filteredCites = [c for c in Cites if hasattr(c, nodeType)]
                    if len(filteredCites) > 1:
                        for n, c1 in enumerate(filteredCites):
                            c1val = getattr(c1, nodeType)
                            c2vals = [getattr(c, nodeType) for c in filteredCites[n:]]
                            if weighted:
                                updateWeightedEdges(tmpgrph, edgeBunchGenerator(c1val, c2vals, weighted = True))
                            else:
                                tmpgrph.add_edges_from(edgeBunchGenerator(c1val, c2vals))
                            if nodeInfo and not hasattr(tmpgrph.node[c1val], 'info'):
                                if nodeType == 'original':
                                    tmpgrph.node[c1val]['info'] = str(c1)
                                elif nodeType == 'journal':
                                    if c1.isJournal():
                                        tmpgrph.node[c1val]['info'] = str(c1.getFullJournalName())
                                    else:
                                        tmpgrph.node[c1val]['info'] = "None"
                                else:
                                    tmpgrph.node[c1val]['info'] = c1val
                                if fullInfo:
                                    tmpgrph.node[c1val]['fullCite'] = str(c1)
                    elif len(filteredCites) == 1:
                        if getattr(filteredCites[0], nodeType) not in tmpgrph and nodeInfo:
                            cval = getattr(filteredCites[0], nodeType)
                            tmpgrph.add_node(cval)
                            if nodeType == 'original':
                                tmpgrph.node[cval]['info'] = str(filteredCites[0])
                            elif nodeType == 'journal':
                                if c1val.isJournal():
                                    tmpgrph.node[cval]['info'] = str(filteredCites[0].getFullJournalName())
                                else:
                                    tmpgrph.node[cval]['info'] = "None"
                            else:
                                tmpgrph.node[cval]['info'] = cval
                            if fullInfo:
                                tmpgrph.node[cval]['fullCite'] = str(filteredCites[0])
                        elif getattr(filteredCites[0], nodeType) not in tmpgrph:
                            tmpgrph.add_node(getattr(filteredCites[0], nodeType))
        else:
            citesSet = set()
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal((count/ len(self)), "Analyzing: " + str(R))
                Cites = R.citations
                if Cites:
                    if dropNonJournals:
                        Cites = filterNonJournals(Cites)
                    if dropAnon:
                        Cites = [c for c in Cites if not c.isAnonymous()]
                    citeHashList = []
                    for cite in Cites:
                        citeHash = hash(cite)
                        if citeHash not in tmpgrph:
                            if cite in citesSet:
                                for c in citesSet:
                                    if cite == c:
                                        citeHash = hash(c)
                            else:
                                citesSet.add(cite)
                                if nodeInfo:
                                    if fullInfo:
                                        tmpgrph.add_node(citeHash, info = str(cite.getID()), fullCite = str(cite))
                                    else:
                                        tmpgrph.add_node(citeHash, info = str(cite.getID()))
                        citeHashList.append(citeHash)
                    if len(citeHashList) > 1:
                        if weighted:
                            for n, c1 in enumerate(citeHashList):
                                updateWeightedEdges(tmpgrph, edgeBunchGenerator(c1, citeHashList[n:], weighted = True))
                        else:
                            for n, c1 in enumerate(citeHashList):
                                tmpgrph.add_edges_from(edgeBunchGenerator(c1, citeHashList[n:]))
        if PBar:
            PBar.finish("Done making a co-citation network of " + repr(self))
        return tmpgrph

    def citationNetwork(self, dropAnon = True, nodeType = "full", extraInfo = False, weighted = True, dropNonJournals = False, saveJournalNames = False):

        """Creates a citation network for the RecordCollection.

        # Parameters

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _authorship_ : `optional [bool]`

        > default `False`, wether to use author's names as the node ID or the whole citations, if `True` names are used if `False` hashes are used

        _extraInfo_ : `optional [bool]`

        > default `True`, wether the original citation string is added to the node as an extra value, if `True` it is

        _weighted_ : `optional [bool]`

        > default `True`, wether the edges are weighted. If `True` the edges are weighted by the number of citations.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, wether to drop ciations of non-journals

        _saveJournalNames_ : `optional [bool]`

        > default `False`, wether to save the full name of the journals of citations. Citations missing a journal will have the string "None", use with _dropNonJournals_ to avoid this

        # Returns

        `Networkx DiGraph`

        > A networkx digraph with hashes as ID and citations as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise ValueError("{} is not an allowed nodeType.".format(nodeType))
        tmpgrph = nx.DiGraph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a citation network")
            count = 0
        else:
            PBar = None
        if nodeType != allowedTypes[0]:
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count/ len(self), "Analyzing: " + str(R))
                reRef = R.createCitation()
                if dropNonJournals:
                    if not reRef.isJournal():
                        continue
                if hasattr(reRef, nodeType):
                    refVal = getattr(reRef, nodeType)
                    if extraInfo and saveJournalNames:
                        tmpgrph.add_node(refVal, info = str(reRef), journal = str( reRef.getFullJournalName()))
                    elif extraInfo:
                        tmpgrph.add_node(refVal, info = str(reRef))
                    elif saveJournalNames:
                        tmpgrph.add_node(refVal, journal = reRef.getFullJournalName())
                    else:
                        tmpgrph.add_node(refVal)
                else:
                    continue
                rCites = R.citations
                if rCites:
                    if dropNonJournals:
                        rCites = filterNonJournals(rCites)
                    if dropAnon:
                        filteredCites = [c for c in rCites if not c.isAnonymous() and hasattr(c, nodeType)]
                    else:
                        filteredCites = [c for c in rCites if hasattr(c, nodeType)]
                    rCitesVals = [getattr(c, nodeType) for c in filteredCites]
                    if extraInfo and saveJournalNames:
                        for i in range(len(filteredCites)):
                            if rCitesVals[i] not in tmpgrph:
                                tmpgrph.add_node(rCitesVals[i], info = str(filteredCites[i]), journal = str(filteredCites[i].getFullJournalName()))
                    elif extraInfo:
                        for i in range(len(filteredCites)):
                            if rCitesVals[i] not in tmpgrph:
                                tmpgrph.add_node(rCitesVals[i], info = str(filteredCites[i]))
                    elif saveJournalNames:
                        for i in range(len(filteredCites)):
                            if rCitesVals[i] not in tmpgrph:
                                tmpgrph.add_node(rCitesVals[i],journal = str(filteredCites[i].getFullJournalName()))
                    if weighted:
                        updateWeightedEdges(tmpgrph, edgeBunchGenerator(refVal, rCitesVals, weighted = True))
                    else:
                        tmpgrph.add_edges_from(edgeBunchGenerator(refVal, rCitesVals))
        else:
            citesSet = set()
            for R in self:
                if PBar:
                    count += 1
                    PBar.updateVal(count / len(self), "Analyzing: " + str(R))
                reRef = R.createCitation()
                if dropNonJournals:
                    if not reRef.isJournal():
                        continue
                recHash = hash(reRef)
                rCites = R.citations
                if dropAnon and reRef.isAnonymous():
                    continue
                if recHash not in tmpgrph:
                    if reRef in citesSet:
                        for c in citesSet:
                            if reRef == c:
                                recHash = hash(c)
                    else:
                        citesSet.add(reRef)
                        if extraInfo and saveJournalNames:
                            tmpgrph.add_node(recHash, info = str(reRef), journal = str(reRef.getFullJournalName()))
                        elif extraInfo:
                            tmpgrph.add_node(recHash, info = str(reRef))
                        elif saveJournalNames:
                            tmpgrph.add_node(recHash, journal = str(reRef.getFullJournalName()))
                        else:
                            tmpgrph.add_node(recHash)
                if rCites:
                    if dropNonJournals:
                        rCites = filterNonJournals(rCites)
                    rCites = [c for c in rCites if not c.isAnonymous()] if dropAnon else rCites
                    citeHashs = []
                    for refCite in rCites:
                        citeHash = hash(refCite)
                        if citeHash not in tmpgrph:
                            if refCite in citesSet:
                                for c in citesSet:
                                    if refCite == c:
                                        citeHash = hash(c)
                            else:
                                citesSet.add(refCite)
                                if extraInfo and saveJournalNames:
                                    tmpgrph.add_node(citeHash, info = str(refCite), journal = str(refCite.getFullJournalName()))
                                elif extraInfo:
                                    tmpgrph.add_node(citeHash, info = str(refCite))
                                elif saveJournalNames:
                                    tmpgrph.add_node(citeHash, journal = str(refCite.getFullJournalName()))
                                else:
                                    tmpgrph.add_node(citeHash)
                        citeHashs.append(citeHash)
                    if weighted:
                        updateWeightedEdges(tmpgrph, edgeBunchGenerator(recHash, citeHashs, weighted = True))
                    else:
                        tmpgrph.add_edges_from(edgeBunchGenerator(recHash, citeHashs))
        if PBar:
            PBar.finish("Done making a citation network of " + repr(self))
        return tmpgrph

    def extractTagged(self, taglist):
        recordsWithTags = set()
        for R in self:
            for t in taglist:
                hasTags = True
                if t not in R.tags:
                    hasTags = False
                    break
            if hasTags:
                recordsWithTags.add(R)
        return RecordCollection(recordsWithTags, repr(self) + "_tags(" + ','.join(taglist) + ')')

    def yearSplit(self, startYear, endYear):
        recordsInRange = set()
        for R in self._Records:
            if R.year >= startYear and R.year <= endYear:
                recordsInRange.add(R)
        return RecordCollection(recordsInRange, repr(self) + "_(" + str(startYear) + " ," + str(endYear) + ")")

    def oneModeNetwork(self, mode, nodeCount = True, edgeWeight = True):
        if mode not in tagsAndNames:
            raise TypeError(str(mode) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a one mode network with " + mode)
            count = 0
        else:
            PBar = None
        grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents = getattr(R, mode, None)
            if contents:
                if isinstance(contents, list):
                    tmplst = [str(n) for n in contents]
                    if len(tmplst) > 1:
                        for i, node1 in enumerate(tmplst):
                            for node2 in tmplst[i + 1:]:
                                if edgeWeight:
                                    try:
                                        grph.edge[node1][node2]['weight'] += 1
                                    except KeyError:
                                        grph.add_edge(node1, node2, weight = 1)
                                else:
                                    if not grph.has_edge(node1, node2):
                                        grph.add_edge(node1, node2)
                            if nodeCount:
                                try:
                                    grph.node[node1]['count'] += 1
                                except KeyError:
                                    grph.node[node1]['count'] = 1
                            else:
                                if not grph.has_node(node1):
                                    grph.add_node(node1)
                    elif len(tmplst) == 1:
                        if nodeCount:
                            try:
                                grph.node[tmplst[0]]['count'] += 1
                            except KeyError:
                                grph.add_node(tmplst[0], count = 1)
                        else:
                            if not grph.has_node(tmplst[0]):
                                grph.add_node(tmplst[0])
                    else:
                        pass
                else:
                    nodeVal = str(contents)
                    if nodeCount:
                        try:
                            grph.node[nodeVal]['count'] += 1
                        except KeyError:
                            grph.add_node(nodeVal, count = 1)
                    else:
                        if not grph.has_node(nodeVal):
                            grph.add_node(nodeVal)
        if PBar:
            PBar.finish("Done making a one mode network with " + mode)
        return grph

    def twoModeNetwork(self, tag1, tag2, directed = False, recordType = True, nodeCount = True, edgeWeight = True):
        if (not tag1 in tagsAndNames) or (not tag2 in tagsAndNames):
            raise TypeError(str(tag1) + " or " + str(tag2) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a two mode network of " + tag1 + " and " + tag2)
            count = 0
        else:
            PBar = None
        if directed:
            grph = nx.DiGraph()
        else:
            grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents1 = getattr(R, tag1, None)
            contents2 = getattr(R, tag2, None)
            if isinstance(contents1, list):
                contents1 = [str(v) for v in contents1]
            elif contents1 == None:
                contents1 = []
            else:
                contents1 = [str(contents1)]
            if isinstance(contents2, list):
                contents2 = [str(v) for v in contents2]
            elif contents2 == None:
                contents2 = []
            else:
                contents2 = [str(contents2)]
            for node1 in contents1:
                for node2 in contents2:
                    if edgeWeight:
                        try:
                            grph.edge[node1][node2]['weight'] += 1
                        except KeyError:
                            grph.add_edge(node1, node2, weight = 1)
                    else:
                        if not grph.has_edge(node1, node2):
                            grph.add_edge(node1, node2)
                if nodeCount:
                    try:
                        grph.node[node1]['count'] += 1
                    except KeyError:
                        try:
                            grph.node[node1]['count'] = 1
                            if recordType:
                                grph.node[node1]['type'] = tag1
                        except KeyError:
                            if recordType:
                                grph.add_node(node1, type = tag1)
                            else:
                                grph.add_node(node1)
                else:
                    if not grph.has_node(node1):
                        if recordType:
                            grph.add_node(node1, type = tag1)
                        else:
                            grph.add_node(node1)
                    elif recordType:
                        if 'type' not in grph.node[node1]:
                            grph.node[node1]['type'] = tag1

            for node2 in contents2:
                if nodeCount:
                    try:
                        grph.node[node2]['count'] += 1
                    except KeyError:
                        try:
                            grph.node[node2]['count'] = 1
                            if recordType:
                                grph.node[node2]['type'] = tag2
                        except KeyError:
                            grph.add_node(node2, count = 1)
                            if recordType:
                                grph.node[node2]['type'] = tag2
                else:
                    if not grph.has_node(node2):
                        if recordType:
                            grph.add_node(node2, type = tag2)
                        else:
                            grph.add_node(node2)
                    elif recordType:
                        if 'type' not in grph.node[node2]:
                            grph.node[node2]['type'] = tag2
        if PBar:
            PBar.finish("Done making a two mode network of " + tag1 + " and " + tag2)
        return grph

    def nModeNetwork(self, tags, recordType = True, nodeCount = True, edgeWeight = True):
        for t in tags:
            if t not in tagsAndNames:
                raise TypeError(str(t) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a " + str(len(tags)) + "-mode network of: " + ', '.join(tags))
            count = 0
        else:
            PBar = None
        grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents = []
            for t in tags:
                tmpVal = getattr(R, t, None)
                if tmpVal:
                    if isinstance(tmpVal, list):
                        contents.append((t, [str(v) for v in tmpVal]))
                    else:
                        contents.append((t, [str(tmpVal)]))
            for i, vlst1 in enumerate(contents):
                for node1 in vlst1[1]:
                    for vlst2 in contents[i + 1:]:
                        for node2 in vlst2[1]:
                            if edgeWeight:
                                try:
                                    grph.edge[node1][node2]['weight'] += 1
                                except KeyError:
                                    grph.add_edge(node1, node2, weight = 1)
                            else:
                                if not grph.has_edge(node1, node2):
                                    grph.add_edge(node1, node2)
                    if nodeCount:
                        try:
                            grph.node[node1]['count'] += 1
                        except KeyError:
                            try:
                                grph.node[node1]['count'] = 1
                                if recordType:
                                    grph.node[node1]['type'] = vlst1[0]
                            except KeyError:
                                if recordType:
                                    grph.add_node(node1, type = vlst1[0])
                                else:
                                    grph.add_node(node1)
                    else:
                        if not grph.has_node(node1):
                            if recordType:
                                grph.add_node(node1, type = vlst1[0])
                            else:
                                grph.add_node(node1)
                        elif recordType:
                            try:
                                grph.node[node1]['type'] += vlst1[0]
                            except KeyError:
                                grph.node[node1]['type'] = vlst1[0]
        if PBar:
            PBar.finish("Done making a " + str(len(tags)) + "-mode network of: " +  ', '.join(tags))
        return grph

    def citeFilter(self, keyString = '', field = 'all', reverse = False, caseSensitive = False):
        """
        Filters Records by some string, keyString, in all of their citations.
        Returns all Records with at least one citation possessing keyString in the field given by field.

        keyString give the string to be searched for if it is is blank then all citations with the specified field will be matched

        field give the component of the citation to be looked at, it is one of a few strings. The default is 'all' which will cause the entire original citation to be searched. It can be used to search across fields, e.g. '1970, V2' is a valid keystring
        The other options are:
        author, searches the author field
        year, searches the year field
        journal, searches the journal field
        V, searches the volume field
        P, searches the page field
        misc, searches all the remaining uncategorized information
        anonymous, searches for anonymous citations, keyString is not used
        bad, searches for bad citations, keyString is not used

        reverse being True causes all Records not matching the query to be returned, default is False

        caseSensitive if True causes the search across the original to be case sensitive, only the 'all' option can be case sensitive

        """
        retRecs = []
        keyString = str(keyString)
        for R in self:
            try:
                if field == 'all':
                    for cite in R.citations:
                        if caseSensitive:
                            if keyString in cite.original:
                                retRecs.append(R)
                                break
                        else:
                            if keyString.upper() in cite.original.upper():
                                retRecs.append(R)
                                break
                elif field == 'author':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.author:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'journal':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.journal:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'year':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.year:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'V':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.V:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'P':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.P:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'misc':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.misc:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'anonymous':
                    for cite in R.citations:
                        if cite.isAnonymous():
                            retRecs.append(R)
                            break
                elif field == 'bad':
                    for cite in R.citations:
                        if cite.bad:
                            retRecs.append(R)
                            break
            except TypeError:
                pass
        if reverse:
            excluded = []
            for R in self:
                if R not in retRecs:
                    excluded.append(R)
            return RecordCollection(inCollection = excluded, name = self._repr + '_subsetByNotCite')
        else:
            return RecordCollection(inCollection = retRecs, name = self._repr + '_subsetByCite')


def isiParser(isifile):
    """
    isiParser() reads the file given by the path isifile, checks that the header is correct then reads until it reaches EF.
    Each it finds is used to initialize a Record then all Record are returned as a list.
    """
    try:
        openfile = open(isifile, 'r', encoding='utf-8-sig')
    except UnicodeDecodeError as e:
        openfile.close()
        raise e
    f = enumerate(openfile, start = 0)
    try:
        linesChecked = 3
        for i in range(linesChecked):
            if "VR 1.0" in f.__next__()[1]:
                break
            if i == linesChecked - 1:
                openfile.close()
                raise BadISIFile(isifile + " Does not have a valid header, 'VR 1.0' not in first two lines")
    except StopIteration as e:
        openfile.close()
        raise BadISIFile("File ends before EF found")
    except UnicodeDecodeError as e:
        openfile.close()
        raise e
    notEnd = True
    plst = []
    while notEnd:
        try:
            line = f.__next__()
        except StopIteration as e:
            raise BadISIFile("File ends before EF found")
        if not line[1]:
            raise BadISIFile("No ER found in " + isifile)
        elif line[1].isspace():
            continue
        elif 'EF' in line[1][:2]:
            notEnd = False
            continue
        else:
            try:
                plst.append(Record(itertools.chain([line], f), isifile, line[0]))
            except BadISIFile as w:
                try:
                    s = f.__next__()[1]
                    while s[:2] != 'ER':
                        s = f.__next__()[1]
                except:
                    raise BadISIFile(str(w) + " could not be resolved")
            except Exception as e:
                openfile.close()
                raise e
    try:
        f.__next__()
        raise BadISIFile("EF not at end of " + isifile)
    except StopIteration as e:
        pass
    finally:
        openfile.close()
    return plst

def getCoCiteIDs(clst):
    """
    Creates a dict of the ID-extra information pairs for a CR tag.
    """
    idDict = {}
    for c in clst:
        cId = c.getID()
        if cId not in idDict:
            idDict[cId] = c.getExtra()
    return idDict

def updateWeightedEdges(grph, ebunch):
    for e in ebunch:
        try:
            grph.edge[e[0]][e[1]]['weight'] += e[2]
        except KeyError:
            grph.add_edge(e[0], e[1], weight = e[2])

def edgeBunchGenerator(base, nodes, weighted = False, reverse = False):
    """
    A helper function for generating a bunch of edges from 1 node base to a list of nodes nodes.
    """
    if weighted and reverse:
        for n in nodes:
            yield (n, base, 1)
    elif weighted:
        for n in nodes:
            yield (base, n, 1)
    elif reverse:
        for n in nodes:
            yield (n, base)
    else:
        for n in nodes:
            yield (base, n)

def edgeNodeReplacerGenerator(base, nodes, loc):
    """
    A helper function for replacing an element of nodes at loc with base
    """
    for n in nodes:
        tmpN = list(n)
        tmpN[loc] = base
        yield tmpN



"""Doc String of tagFuncs"""

from .constants import tagNameConverter, monthDict
from .citation import Citation

def pubType(val):
    """
    returns the type of publication as a character: conference, book, journal, book in series, or patent

    PT tag
    """
    return val[0]


def authorsFull(val):
    """
    returns a list of authors full names

    AF tag
    """
    return val

def group(val):
    """
    returns the group associated with the Record

    GP tag
    """
    return val[0]

def editedBy(val):
    """
    returns a list of the editors of the Record

    BE tag
    """
    return val

def authorsShort(val):
    """
    returns a list of authors shortened names

    AU tag
    """
    return val

def bookAuthor(val):
    """
    returns a list of the short names of the authors of a book Record

    BA tag
    """
    return val

def bookAuthorFull(val):
    """
    returns a list of the long names of the authors of a book Record

    BF tag
    """
    return val

def groupName(val):
    """
    returns the name of the group associated with the Record

    CA tag
    """
    return val[0]

def title(val):
    """
    returns the title of the record

    TI tag
    """
    return ' '.join(val)

def editors(val):
    """
    # Needs Work
    """
    return val

def journal(val):
    """
    returns the full name of the publication

    SO tag
    """
    return ' '.join(val)

def seriesTitle(val):
    """
    returns the title of the series the Record is in

    SE tag
    """
    return val[0]

def seriesSubtitle(val):
    """
    returns the title of the series the Record is in

    BS tag
    """
    return val[0]

def language(val):
    """
    returns the languages of the Record as a string with languages seperated by ', ', usually there is only one language

    LA tag
    """
    return ', '.join(val)

def docType(val):
    """
    returns the type of document the Record contains

    DT tag
    """
    return val[0]

def confTitle(val):
    """
    returns the title of the conference associated with the Record

    CT tag
    """
    return ' '.join(val)

def confDate(val):
    """
    returns the date string of the conference associated with the Record

    CY tag
    """
    return val[0]

def confSponsors(val):
    """
    returns a list of sponsors for the conference associated with the record

    SP tag
    """
    return ', '.join(val).split(', ')

def wosTimesCited(val):
    """
    returns the number of times the Record has been cited byr records in WOS

    TC tag
    """
    return int(val[0])

def authAddress(val):
    """

    C1 tag
    """
    ret = []
    for a in val:
        if a[0] == '[':
            ret.append('] '.join(a.split('] ')[1:]))
        else:
            ret.append(a)
    return ret

def confLocation(val):
    """
    returns the sting giving the confrence's location

    CL tag
    """
    return ' '.join(val)

def j9(val):
    """
    returns the J9 (29-Character Source Abbreviation) of the publication

    J9 tag
    """
    return val[0]

def funding(val):
    """
    Returns a list of the groups funding the Record

    FU tag
    """
    return ' '.join(val).split('; ')

def subjectCategory(val):
    """
    returns a list of the subjects associated with the Record

    SC tag
    """
    return ' '.join(val).split('; ')

def citations(val):
    """
    returns a list of all the citations in the record

    CR tag
    """
    retCites = []
    for c in val:
        retCites.append(Citation(c))
    return retCites

def publisherCity(val):
    """
    Returns the city the publisher is in

    PI tag
    """
    return ' '.join(val).upper()

def ISSN(val):
    """
    returns the ISSN of the Record

    SN tag
    """
    return val[0]

def articleNumber(val):
    """
    returns a string giving the article number, not all are integers

    AR tag
    """
    return val[0]

def issue(val):
    """
    returns a string giving the issue or range of issues the Record was in

    IS tag
    """
    return val[0]

def email(val):
    """
    returns a list of emails given by the authors of the Record

    EM tag
    """
    return ' '.join(val).split('; ')

def eISSN(val):
    """
    returns the EISSN of the Record

    EI tag
    """
    return val[0]

def DOI(val):
    """
    return the DOI number of the record

    DI tag
    """
    return val[0]

def wosString(val):
    """
    returns the WOS number of the record as a string preceded by "WOS:"

    UT tag
    """
    return val[0]

def orcID(val):
    """
    returns a list of orc IDs of the Record

    OI tag
    """
    return ' '.join(val).split('; ')

def meetingAbstract(val):
    """
    returns the ID of the meeting abstract prefixed by 'EPA-'

    MA tag
    """
    return val[0]

def isoAbbreviation(val):
    """
    returns the iso abbreviation of the journal

    JI tag
    """
    return ' '.join(val).upper()

def pageCount(val):
    """
    returns an interger giving the number of pages of the Record

    PG tag
    """
    return int(val[0])

def publisher(val):
    """
    returns the publisher of the Record

    PU tag
    """
    return ' '.join(val).upper()

def ISBN(val):
    """
    returns a list of ISBNs assocaited with the Record

    BN tag
    """
    return ' '.join(val).split('; ')

def month(val):
    """
    returns the month the record was published in as an int with January as 1, February 2, ...

    PD tag
    """
    return getMonth(val[0])

def fundingText(val):
    """
    Returns a string of the funding thanks

    FX tag
    """
    return ' '.join(val)

def bookDOI(val):
    """
    returns the book DOI of the Record

    D2 tag
    """
    return val[0]

def volume(val):
    """
    return the volume the record is in as a string not an int

    VL tag
    """
    return val[0].strip()

def ResearcherIDnumber(val):
    """
    returns a lsit of the research ids of the Record

    RI tag
    """
    return ' '.join(val).split('; ')

def citedRefsCount(val):
    """
    returns the numer citations, length of CR list

    NR tag
    """
    return int(val[0])

def beginningPage(val):
    """
    returns the first page the record occurs on as a string not an int

    BP tag
    """
    return val[0].strip()

def abstract(val):
    """
    return abstract of the record, with newlines hopefully in the correct places

    AB tag
    """
    return '\n'.join(val)

def supplement(val):
    """
    returns the supplemtn number

    SU tag
    """
    return ' '.join(val)

def confHost(val):
    """
    returns the host of the conference

    HO tag
    """
    return ' '.join(val)

def publisherAddress(val):
    """
    returns the publishers address

    PA tag
    """
    return ' '.join(val)

def endingPage(val):
    """
    return the last page the record occurs on as a string not an int

    EP tag
    """
    return val[0].strip()

def year(val):
    """
    returns the year the record was published in as an int

    PY tag
    """
    yearField = val[0]
    if len(yearField) == 4:
        return int(yearField)
    else:
        raise ValueError("Incorrectly formatted PY tag")

def authKeyWords(val):
    """
    returns the keywords assigned by the author of the Record

    DE tag
    """
    return ' '.join(val).split('; ')

def reprintAddress(val):
    """
    returns the reprint address string

    RP tag
    """
    return val[0]

def totalTimesCited(val):
    """
    returns the total number of citations of the record

    Z9 tag
    """
    return int(val[0])

def partNumber(val):
    """
    return an integer giving the part of the issue the Record is in

    PN tag
    """
    return int(val[0])

def specialIssue(val):
    """
    returns the special issue value

    SI tag
    """
    return val[0]

def subjects(val):
    """
    returns a lsit of subjects as assigned by WOS

    WC tag
    """
    return ' '.join(val).split('; ')

def keyWords(val):
    """
    returns the WOS keywords of the Record

    ID tag
    """
    return ' '.join(val).split('; ')

def pubMedID(val):
    """
    returns the pubmed idof the record

    PM tag
    """
    return val[0]

def documentDeliveryNumber(val):
    """
    returns the document delivery number of the Record

    GA tag
    """
    return val[0]

def makeReversed(d):
    """
    Simple function for reversing a dictionary
    """
    dTmp = d.copy()
    for k in d.keys():
        try:
            dTmp[tagNameConverter[k]] = dTmp[k]
        except KeyError:
            raise Exception("Something is wrong with the tag to full name mappings")
    return dTmp

tagToFunc = makeReversed( {
            'PT' : pubType,
            'AF' : authorsFull,
            'GP' : group,
            'BE' : editedBy,
            'AU' : authorsShort,
            'BA' : bookAuthor,
            'BF' : bookAuthorFull,
            'CA' : groupName,
            'ED' : editors,
            'TI' : title,
            'SO' : journal,
            'SE' : seriesTitle,
            'BS' : seriesSubtitle,
            'LA' : language,
            'DT' : docType,
            'CT' : confTitle,
            'CY' : confDate,
            'HO' : confHost,
            'CL' : confLocation,
            'SP' : confSponsors,
            'DE' : authKeyWords,
            'ID' : keyWords,
            'AB' : abstract,
            'C1' : authAddress,
            'RP' : reprintAddress,
            'EM' : email,
            'RI' : ResearcherIDnumber,
            'OI' : orcID,
            'FU' : funding,
            'FX' : fundingText,
            'CR' : citations,
            'NR' : citedRefsCount,
            'TC' : wosTimesCited,
            'Z9' : totalTimesCited,
            'PU' : publisher,
            'PI' : publisherCity,
            'PA' : publisherAddress,
            'SC' : subjectCategory,
            'SN' : ISSN,
            'EI' : eISSN,
            'BN' : ISBN,
            'J9' : j9,
            'JI' : isoAbbreviation,
            'PD' : month,
            'PY' : year,
            'VL' : volume,
            'IS' : issue,
            'PN' : partNumber,
            'SU' : supplement,
            'SI' : specialIssue,
            'MA' : meetingAbstract,
            'BP' : beginningPage,
            'EP' : endingPage,
            'AR' : articleNumber,
            'PG' : pageCount,
            'WC' : subjects,
            'DI' : DOI,
            'D2' : bookDOI,
            'GA' : documentDeliveryNumber,
            'UT' : wosString,
            'PM' : pubMedID,
            })

def getMonth(s):
    """
    Known formats:
    Month ("%b")
    Month Day ("%b %d")
    Month-Month ("%b-%b") --- this gets coerced to the first %b, dropping the month range
    Season ("%s") --- this gets coerced to use the first month of the given season
    Month Day Year ("%b %d %Y")
    Month Year ("%b %Y")
    """
    monthOrSeason = s.split(' ')[0].split('-')[0].upper()
    if monthOrSeason in monthDict:
        return monthDict[monthOrSeason]
    else:
        raise ValueError("Month format not recognized: " + s)



import unittest
import metaknowledge
import os
import io
from metaknowledge.graphHelpers import _ProgressBar

fileShortName = 'testNetworks'
fileEName = 'testNetworks_edgeList_undirected.tst'
fileNName = 'testNetworks_nodeAttributes_undirected.tst'
filesuffix = 'tst'

class TestHelpers(unittest.TestCase):
    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")
        self.G = self.RC.coCiteNetwork()
    def test_graphwrite(self):
        metaknowledge.write_graph(self.G, fileShortName, suffix = filesuffix)
        tmpG = metaknowledge.read_graph(fileEName, fileNName)
        self.assertEqual(len(tmpG.edges()), len(self.G.edges()))
        self.assertEqual(len(tmpG.nodes()), len(self.G.nodes()))
        os.remove(fileEName)
        os.remove(fileNName)

    def test_progress(self):
        metaknowledge.VERBOSE_MODE = True
        tmpIO = io.StringIO()
        P = _ProgressBar(0, "testing", output = tmpIO)
        metaknowledge.write_edgeList(self.G, fileEName, progBar = P)
        tmpIO.seek(0)
        s = ''.join(tmpIO.readlines())
        self.assertFalse(
        "==================================================]100.0%" in s)
        self.assertTrue("Writing edge list testNetworks_edgeList_undirected.tst" in s)
        os.remove(fileEName)
        metaknowledge.VERBOSE_MODE = False
    """
    def test_dropEdges(self):
        tmpG = metaknowledge.drop_edges(self.G, 4)
        self.assertEqual(len(tmpG.edges()), len(self.G.edges()))

    def test_dropNodes(self):
        tmpG = metaknowledge.drop_nodes(self.G, 4)
        self.assertEqual(len(tmpG.nodes()), len(self.G.nodes()))
    """



from .journalAbbreviations import getj9dict, manaulDBname, addToDB

abbrevDict = None

class BadCitation(Warning):
    """
    Exception thrown by Citation
    """
    pass

class Citation(object):
    """
    A class to hold citation strings and allow for comparison between them.

    The initializer takes in a string representing a WOS citation they are in the form:

    > Author, Year, Journal, Volume, Page, DOI

    Author is the author's name in the form of first last name first initial sometimes followed by a period.
    Year is the year of publication.
    Journal being the 29-Character Source Abbreviation of the journal.
    Volume is the volume number(s) of the publication preceded by a V
    Page is the page number the record starts on
    DOI is the DOI number of the cited record preceeded by the letters "DOI"
    Combined they look like:

    > Nunez R., 1998, MATH COGNITION, V4, P85, DOI 10.1080/135467998387343

    Note that any of the fields have been known to be missing and the requirements for the fields are not always met. If something is in the source string that cannot be interpeted as any of these it is put in the `misc` attribute.

    The reason for this class is that the WOS data are often irregular. It is designed to allow comparison between WOS citation strings, even when they are missing pieces.

    # Customizations

    Citation's hashing and equality checking are based on what data they have. The equality checking first checks both Citation's DOI's and if either is missing moves to the other fields. If any of the fields disagree `False` is returned (note, authors are not compared if one is anonymous) if they all agree, including the `misc` field, then True is returned.

    Unfortunately this type of equality checking precludes hashes being identical so to compare Citation objects always use ==. Hashes, if identical, indicate the Citations are identical (excluding collisions), but the converse is not True.

    When converted to a string a Citation will return the original string.

    # Attributes

    As noted above, citations are considered to be divided into six distinct fields (Author, Year, Journal, Volume, Page and DOI) with a seventh misc for anything not in those. Records thus have an attribute with a name corresponding to each `author`, `year`, `journal`, `V`, `P`, `DOI` and `misc` respectively. These are created if there is anything in the field. So a Citation created from the string: "Nunez R., 1998, MATH COGNITION" would have `author`, `year` and `journal` defined. While one from "Nunez R." would have only the attribute `misc`.

    If the parsing of a citation string fails the attribute `bad` is set to True and the attribute `error` is created to contain the error, which is a [BadCitation](#metaknowledge.BadCitation) object. If no errors occur `bad` is `False`.

    The attribute `original` is the unmodified string (_cite_) given to create the Citation, it can also be accessed by converting to a string, e.g. with `str()`.

    # \_\_Init\_\_

    Citations can be created by [Records](#metaknowledge.Record) or by giving the initializer a string containing a WOS style citation.

    # Parameters

    _cite_ : `str`

    > A str containing a WOS style citation.
    """
    def __init__(self, cite):
        self.original = cite
        self.bad = False
        self._isjourn = None
        c = ' '.join(cite.upper().split()).split(', ')
        if 'DOI' in c[-1][:3]:
            self.DOI = c.pop().split(' ')[-1]
        if len(c) < 3:
            self.bad = True
            self.error = BadCitation("Too few elements")
            if len(c) == 2:
                self.author = c.pop(0).replace('.','')
            self.misc = ', '.join(c)
        else:
            if not c[0].isnumeric():
                self.author = c.pop(0).replace('.','')
            if c[0].isnumeric():
                self.year = c.pop(0)
            if not c[0].isnumeric():
                self.journal = c.pop(0)
            else:
                self.misc = c.pop(0)
                self.bad = True
                self.error = BadCitation("Too many numbers")
            for field in c:
                if 'V' == field[0] and field[1:].isnumeric():
                    self.V = field
                elif 'P' == field[0] and field[1:].isnumeric():
                    self.P = field
                else:
                    if hasattr(self, 'misc'):
                        self.misc += ', ' + field
                    else:
                        self.misc = field

    def __str__(self):
        """
        returns the original string
        """
        return self.original

    def __hash__(self):
        """
        A hash for Citation that should be equal to the hash of other citations that are equal to it
        """
        if hasattr(self, 'DOI'):
            return hash(self.DOI)
        elif self.bad and getattr(self, 'misc', False):
            return hash(getattr(self, 'misc'))
        else:
            hashedString = getattr(self, 'author', '')
            hashedString += getattr(self, 'year', '')
            hashedString += getattr(self, 'journal', '')
            hashedString += getattr(self, 'V', '')
            hashedString += getattr(self, 'P', '')
            return hash(hashedString)

    def __eq__(self, other):
        """
        First checks DOI for equality then checks each attribute if any are not equal False is returned
        """
        if not isinstance(other, Citation):
            return NotImplemented
        elif getattr(self, 'DOI', None) == getattr(other, 'DOI', False):
            return True
        elif getattr(self, 'author', None) != getattr(other, 'author', None) and getattr(self, 'author', None) != '[ANONYMOUS]' and getattr(other, 'author', False) != '[ANONYMOUS]':
            return False
        elif getattr(self, 'year', None) != getattr(other, 'year', None):
            return False
        elif  getattr(self, 'journal', None) != getattr(other, 'journal', None):
            return False
        elif getattr(self, 'V', None) and getattr(other, 'V', False) and getattr(self, 'V', None) != getattr(other, 'V', False):
            return False
        elif getattr(self, 'P', False) and getattr(other, 'P', False) and getattr(self, 'P', None) != getattr(other, 'P', False):
            return False
        elif self.bad and other.bad and not getattr(self, 'misc', None) == getattr(other, 'misc', None):
            return False
        else:
            return True

    def __ne__(self, other):
        """
        Returns the inverse of equality
        """
        return not self == other

    def __repr__(self):
        """
        the representation of the Citation is its original form
        """
        return self.original

    def isAnonymous(self):
        """
        Checks if the author is given as "[ANONYMOUS]" and returns `True` if so.

        # Returns

        `bool`

        > True if the author is ANONYMOUS otherwise `False`.
        """
        if hasattr(self, 'author'):
            return self.author == "[ANONYMOUS]"
        else:
            return False

    def getID(self):
        """
        Returns "author, year, journal" if available and "misc" otherwise. It is for shortening labels when creating networks as the resultant strings are often unique. [`getExtra()`](#Citation.getExtra) gets everthing not returned by `getID()`.

        # Returns

        `str`

        > A string to use as the shortened ID of a node.
        """
        if hasattr(self, 'author') or hasattr(self, 'journal'):
            retid = ''
            if hasattr(self, 'author'):
                retid += self.author
            if hasattr(self, 'year'):
                if retid == '':
                    retid += self.year
                else:
                    retid += ', '  + self.year
            if hasattr(self, 'journal'):
                if retid == '':
                    retid += self.journal
                else:
                    retid += ', '  + self.journal
            return retid
        else:
            return self.misc

    def getExtra(self):
        """
        Returns any V, P or misc values as a string. These are all the values not returned by [`getID()`](#Citation.getID).

        # Returns

        `str`

        > A string containing the data not in the ID of the Citation.
         """
        extraTags = ['journal','V', 'P', 'misc']
        retVal = ""
        for tag in extraTags:
            if hasattr(self, tag):
                retVal += getattr(self, tag) + ', '
        if len(retVal) > 2:
            return retVal[:-2]
        else:
            return retVal

    def isJournal(self, manaulDB = manaulDBname, returnDict = 'both', checkIfExcluded = False):
        """Returns `True` if the Citation's journal field is a journal abbreviation given by WOS, i.e. checks if the citation is citing a journal. Requires the j9Abbreviations database file.

        # Returns

        `bool`

        > `True` if the Citation is for a journal
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict(manualDB = manaulDB, returnDict = returnDict)
        if checkIfExcluded and hasattr(self, 'journal'):
            try:
                if abbrevDict.get( self.journal, [True])[0]:
                    return False
                else:
                    return True
            except IndexError:
                return False
        elif self._isjourn is None:
            if hasattr(self, 'journal'):
                dictVal = abbrevDict.get(self.journal, [b''])[0]
                if dictVal:
                    self._isjourn = dictVal
                else:
                    self._isjourn = False
            else:
                self._isjourn = False
        return self._isjourn

    def getFullJournalName(self):
        """Returns the full name of the Citation's journal field. Requires the j9Abbreviations database file.

        # Returns

        `str`

        > The first full name given for the journal of the Citation (or the first name in the WOS list if multiple names exist), if there is not one then `None` is returned
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict()
        if self.isJournal():
            return abbrevDict[self.journal][0]
        else:
            return None

    def addToDB(self, manualName = None, manaulDB = manaulDBname, invert = False):
        try:
            if invert:
                d = {self.journal : ''}
            elif manualName is None:
                d = {self.journal : self.journal}
            else:
                d = {self.journal : manualName}
            addToDB(abbr = d, dbname = manaulDB)
        except KeyError:
            raise KeyError("This citation does not have a journal field.")
        else:
            global abbrevDict
            abbrevDict.update(d)

def filterNonJournals(citesLst, invert = False):
    retCites = []
    for c in citesLst:
        if c.isJournal():
            if not invert:
                retCites.append(c)
        elif invert:
            retCites.append(c)
    return retCites



import unittest
import metaknowledge

class TestRecord(unittest.TestCase):
    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.R = metaknowledge.Record(simplePaperString)
        self.Rbad = metaknowledge.Record(simplePaperString[:-3])
    def test_isRecord(self):
        self.assertTrue(isinstance(self.R, metaknowledge.Record))
        metaknowledge.btest(quite = True)
    def test_bad(self):
        self.assertTrue(self.Rbad.bad)
        with self.assertRaises(TypeError):
            metaknowledge.Record(set('a','b'))
    def test_equality(self):
        self.assertEqual(self.R, self.R)
        self.assertTrue(self.R != self.Rbad)
    def test_hash(self):
        self.assertNotEqual(hash(self.R), hash(self.Rbad))
    def test_state(self):
        state = self.R.__getstate__()
        Rtmp = metaknowledge.Record('PT J')
        Rtmp.__setstate__(state)
        self.assertEqual(self.R, Rtmp)
    def test_title(self):
        self.assertEqual(self.R.title, "Example Paper")
        self.assertEqual(str(self.R), "Example Paper")
    def test_author(self):
        self.assertEqual(self.R.authorsFull, ["John, Doe"])
    def test_year(self):
        self.assertEqual(self.R.year, 2015)
    def test_month(self):
        self.assertEqual(self.R.month, 4)
    def test_cites(self):
        self.assertEqual(str(self.R.citations[0]), "John D. 1999, TOPICS IN COGNITIVE SCIENCE")
        self.assertEqual(len(self.R.citations), 1)
    def test_WOS(self):
        self.assertEqual(self.R.wosString, 'WOS:123317623000007')
    def test_citationGen(self):
        self.assertTrue(self.R.createCitation() == metaknowledge.Citation("John, Doe, 2015, EXAMPLE, V1, P1, DOI 10.1111"))
    def test_journal(self):
        self.assertEqual(self.R.journal, 'TOPICS IN COGNITIVE SCIENCE')
    def test_tags(self):
        self.assertEqual(self.R.getTagsDict(['a', 'b']), {'a': None, 'b': None})

simplePaperString = """PT J
AU John, D
AF John, Doe
TI Example Paper
SO TOPICS IN COGNITIVE SCIENCE
LA English
DT Article
DE Example; testing
ID REAL; TIME
AB This is a test.
C1 UW, Ontario, Canada.
RP John, D (reprint author), UW, Ontario, Canada.
CR John D. 1999, TOPICS IN COGNITIVE SCIENCE
J9 EXAMPLE
JI examaple
PD APR
PY 2015
VL 1
BP 1
EP 2
DI 10.1111
UT WOS:123317623000007
ER
"""
